"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from typing import (
    NoReturn,
)

from minos.common import (
    PostgreSqlMinosDatabase,
)


class HandlerSetup(PostgreSqlMinosDatabase):
    """Minos Broker Setup Class"""

    TABLE_NAME: str

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def _setup(self) -> NoReturn:
        await self._create_event_queue_table()

    async def _create_event_queue_table(self) -> NoReturn:
        await self.submit_query(
            'CREATE TABLE IF NOT EXISTS "%s" ('
            '"id" BIGSERIAL NOT NULL PRIMARY KEY, '
            '"topic" VARCHAR(255) NOT NULL, '
            '"partition_id" INTEGER,'
            '"binary_data" BYTEA NOT NULL, '
            '"creation_date" TIMESTAMP NOT NULL);' % (self.TABLE_NAME)
        )



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

from __future__ import (
    annotations,
)

from datetime import (
    datetime,
)
from typing import (
    Iterable,
    Optional,
    Type,
    Union,
)

from minos.common import (
    Aggregate,
    import_module,
)


class SnapshotEntry(object):
    """Minos Snapshot Entry class.

    Is the python object representation of a row in the ``snapshot`` storage system.
    """

    __slots__ = "aggregate_id", "aggregate_name", "version", "data", "created_at", "updated_at"

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        aggregate_id: int,
        aggregate_name: str,
        version: int,
        data: Union[bytes, memoryview] = bytes(),
        created_at: Optional[datetime] = None,
        updated_at: Optional[datetime] = None,
    ):
        if isinstance(data, memoryview):
            data = data.tobytes()

        self.aggregate_id = aggregate_id
        self.aggregate_name = aggregate_name
        self.version = version
        self.data = data

        self.created_at = created_at
        self.updated_at = updated_at

    @classmethod
    def from_aggregate(cls, aggregate: Aggregate) -> SnapshotEntry:
        """Build a new instance from an ``Aggregate``.

        :param aggregate: The aggregate instance.
        :return: A new ``MinosSnapshotEntry`` instance.
        """
        # noinspection PyTypeChecker
        return cls(aggregate.id, aggregate.classname, aggregate.version, aggregate.avro_bytes)

    @property
    def aggregate(self) -> Aggregate:
        """Rebuild the stored ``Aggregate`` object instance from the internal state.

        :return: A ``Aggregate`` instance.
        """
        cls = self.aggregate_cls
        instance = cls.from_avro_bytes(self.data, id=self.aggregate_id, version=self.version)
        return instance

    @property
    def aggregate_cls(self) -> Type[Aggregate]:
        """Load the concrete ``Aggregate`` class.

        :return: A ``Type`` object.
        """
        # noinspection PyTypeChecker
        return import_module(self.aggregate_name)

    def __eq__(self, other: SnapshotEntry) -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __hash__(self) -> int:
        return hash(tuple(self))

    def __iter__(self) -> Iterable:
        # noinspection PyRedundantParentheses
        yield from (self.aggregate_name, self.version, self.data, self.created_at, self.updated_at)

    def __repr__(self):
        name = type(self).__name__
        return (
            f"{name}(aggregate_id={repr(self.aggregate_id)}, aggregate_name={repr(self.aggregate_name)}, "
            f"version={repr(self.version)}, data={repr(self.data)}, "
            f"created_at={repr(self.created_at)}, updated_at={repr(self.updated_at)})"
        )



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

from typing import (
    Any,
)

from aiomisc.service.periodic import (
    PeriodicService,
    Service,
)

from .consumers import (
    CommandReplyConsumer,
)
from .handlers import (
    CommandReplyHandler,
)


class CommandReplyConsumerService(Service):
    """Minos QueueDispatcherService class."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dispatcher = CommandReplyConsumer.from_config(**kwargs)

    async def start(self) -> None:
        """Method to be called at the startup by the internal ``aiomisc`` loigc.

        :return: This method does not return anything.
        """
        await self.dispatcher.setup()
        await self.dispatcher.dispatch()

    async def stop(self, exception: Exception = None) -> Any:
        """Stop the service execution.

        :param exception: Optional exception that stopped the execution.
        :return: This method does not return anything.
        """
        await self.dispatcher.destroy()


class CommandReplyHandlerService(PeriodicService):
    """Minos QueueDispatcherService class."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dispatcher = CommandReplyHandler.from_config(**kwargs)

    async def start(self) -> None:
        """Method to be called at the startup by the internal ``aiomisc`` loigc.

        :return: This method does not return anything.
        """
        await self.dispatcher.setup()
        await super().start()

    async def callback(self) -> None:
        """Method to be called periodically by the internal ``aiomisc`` logic.

        :return:This method does not return anything.
        """
        await self.dispatcher.dispatch()

    async def stop(self, err: Exception = None) -> None:
        """Stop the service execution.

        :param err: Optional exception that stopped the execution.
        :return: This method does not return anything.
        """
        await super().stop(err)
        await self.dispatcher.destroy()



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.

from __future__ import (
    annotations,
)

from minos.common import (
    Event,
    MinosConfig,
)

from ..abc import (
    Consumer,
)


class EventConsumer(Consumer):
    """Event Consumer class."""

    TABLE_NAME = "event_queue"

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> EventConsumer:
        topics = [item.name for item in config.events.items]
        return cls(topics=topics, broker=config.events.broker, **config.events.queue._asdict(), **kwargs)

    def _is_valid_instance(self, value: bytes):
        try:
            Event.from_avro_bytes(value)
            return True
        except:  # noqa E722
            return False



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
__version__ = "0.0.5"

from .brokers import (
    CommandBroker,
    CommandReplyBroker,
    EventBroker,
    Producer,
    ProducerService,
)
from .exceptions import (
    MinosNetworkException,
    MinosPreviousVersionSnapshotException,
    MinosSnapshotException,
)
from .handlers import (
    CommandConsumer,
    CommandConsumerService,
    CommandHandler,
    CommandHandlerService,
    CommandReplyConsumer,
    CommandReplyConsumerService,
    CommandReplyHandler,
    CommandReplyHandlerService,
    Consumer,
    EventConsumer,
    EventConsumerService,
    EventHandler,
    EventHandlerService,
    Handler,
    HandlerSetup,
)
from .rest import (
    RestBuilder,
    RestService,
)
from .snapshots import (
    SnapshotBuilder,
    SnapshotEntry,
    SnapshotService,
)



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.
from __future__ import (
    annotations,
)

import datetime
from abc import (
    abstractmethod,
)
from typing import (
    Any,
    NoReturn,
    Optional,
)

from aiokafka import (
    AIOKafkaConsumer,
)
from psycopg2.extensions import (
    AsIs,
)

from minos.common import (
    BROKER,
)

from .setups import (
    HandlerSetup,
)


class Consumer(HandlerSetup):
    """
    Handler Server

    Generic insert for queue_* table. (Support Command, CommandReply and Event)

    """

    __slots__ = "_topics", "_broker", "__consumer"

    def __init__(self, *, topics: list[str], broker: Optional[BROKER] = None, consumer: Optional[Any] = None, **kwargs):
        super().__init__(**kwargs)
        self._topics = topics
        self._broker = broker
        self.__consumer = consumer

    async def _setup(self) -> NoReturn:
        await super()._setup()
        await self._consumer.start()

    @property
    def _consumer(self) -> AIOKafkaConsumer:
        if self.__consumer is None:  # pragma: no cover
            self.__consumer = AIOKafkaConsumer(
                *self._topics, bootstrap_servers=f"{self._broker.host}:{self._broker.port}"
            )
        return self.__consumer

    async def _destroy(self) -> NoReturn:
        await self._consumer.stop()
        await super()._destroy()

    async def dispatch(self) -> NoReturn:
        """Perform a dispatching step.

        :return: This method does not return anything.
        """
        await self.handle_message(self._consumer)

    async def handle_message(self, consumer: Any) -> NoReturn:
        """Message consumer.

        It consumes the messages and sends them for processing.

        Args:
            consumer: Kafka Consumer instance (at the moment only Kafka consumer is supported).
        """

        async for msg in consumer:
            await self.handle_single_message(msg)

    async def handle_single_message(self, msg):
        """Handle Kafka messages.

        Evaluate if the binary of message is an Event instance.
        Add Event instance to the event_queue table.

        Args:
            msg: Kafka message.

        Raises:
            Exception: An error occurred inserting record.
        """
        # the handler receive a message and store in the queue database
        # check if the event binary string is well formatted
        if not self._is_valid_instance(msg.value):
            return

        return await self.queue_add(msg.topic, msg.partition, msg.value)

    @abstractmethod
    def _is_valid_instance(self, value: bytes):  # pragma: no cover
        raise Exception("Method not implemented")

    async def queue_add(self, topic: str, partition: int, binary: bytes) -> int:
        """Insert row to event_queue table.

        Retrieves number of affected rows and row ID.

        Args:
            topic: Kafka topic. Example: "TicketAdded"
            partition: Kafka partition number.
            binary: Event Model in bytes.

        Returns:
            Queue ID.

            Example: 12

        Raises:
            Exception: An error occurred inserting record.
        """
        queue_id = await self.submit_query_and_fetchone(
            _INSERT_QUERY, (AsIs(self.TABLE_NAME), topic, partition, binary, datetime.datetime.now()),
        )

        return queue_id[0]


_INSERT_QUERY = """
INSERT INTO %s (topic, partition_id, binary_data, creation_date)
VALUES (%s, %s, %s, %s)
RETURNING id;
""".strip()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from collections import (
    namedtuple,
)
from datetime import (
    datetime,
)
from typing import (
    Callable,
)

from minos.common import (
    MinosModel,
)

HandlerEntry = namedtuple(
    "HandlerEntry",
    {"id": int, "topic": str, "callback": Callable, "partition_id": int, "data": MinosModel, "created_at": datetime},
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

from typing import (
    Optional,
)

from minos.common import (
    Aggregate,
    Event,
    MinosConfig,
)

from .abc import (
    Broker,
)


class EventBroker(Broker):
    """Minos Event broker class."""

    ACTION = "event"

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> EventBroker:
        return cls(*args, **config.events.queue._asdict(), **kwargs)

    async def send(self, items: list[Aggregate], topic: Optional[str] = None, **kwargs) -> int:
        """Send a list of ``Aggregate`` instances.

        :param items: A list of aggregates.
        :param topic: Topic in which the message will be published.
        :return: This method does not return anything.
        """
        if topic is None:
            topic = self.topic
        event = Event(topic, items)
        return await self._send_bytes(event.topic, event.avro_bytes)



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.
from __future__ import (
    annotations,
)

from minos.common import (
    CommandReply,
    MinosConfig,
)

from ..abc import (
    Consumer,
)


class CommandReplyConsumer(Consumer):
    """Command Reply consumer class."""

    TABLE_NAME = "command_reply_queue"

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> CommandReplyConsumer:
        topics = [f"{item.name}Reply" for item in config.saga.items]
        return cls(topics=topics, broker=config.saga.broker, **config.saga.queue._asdict(), **kwargs)

    def _is_valid_instance(self, value: bytes) -> bool:
        try:
            CommandReply.from_avro_bytes(value)
            return True
        except:  # noqa E722
            return False



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from .abc import (
    Broker,
)
from .command_replies import (
    CommandReplyBroker,
)
from .commands import (
    CommandBroker,
)
from .events import (
    EventBroker,
)
from .producers import (
    Producer,
)
from .services import (
    ProducerService,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from .consumers import (
    Consumer,
)
from .handlers import (
    Handler,
)
from .setups import (
    HandlerSetup,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

from .consumers import (
    CommandReplyConsumer,
)
from .handlers import (
    CommandReplyHandler,
)
from .services import (
    CommandReplyConsumerService,
    CommandReplyHandlerService,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

from typing import (
    Optional,
)

from minos.common import (
    Command,
    MinosConfig,
    MinosModel,
)

from .abc import (
    Broker,
)


class CommandBroker(Broker):
    """Minos Command Broker Class."""

    ACTION = "command"

    def __init__(self, *args, saga_uuid: Optional[str] = None, reply_topic: Optional[str] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.reply_topic = reply_topic
        self.saga_uuid = saga_uuid

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> CommandBroker:
        return cls(*args, **config.commands.queue._asdict(), **kwargs)

    async def send(
        self,
        items: list[MinosModel],
        topic: Optional[str] = None,
        saga_uuid: Optional[str] = None,
        reply_topic: Optional[str] = None,
        **kwargs
    ) -> int:
        """Send a list of ``Aggregate`` instances.

        :param items: A list of aggregates.
        :param topic: Topic in which the message will be published.
        :param saga_uuid: Saga identifier.
        :param reply_topic: Topic name in which the reply will be published.
        :return: This method does not return anything.
        """
        if topic is None:
            topic = self.topic
        if saga_uuid is None:
            saga_uuid = self.saga_uuid
        if reply_topic is None:
            reply_topic = self.reply_topic
        command = Command(topic, items, saga_uuid, reply_topic)
        return await self._send_bytes(command.topic, command.avro_bytes)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from minos.common import (
    Aggregate,
    MinosException,
)


class MinosNetworkException(MinosException):
    """Base network exception."""


class MinosSnapshotException(MinosNetworkException):
    """Base snapshot exception"""


class MinosPreviousVersionSnapshotException(MinosSnapshotException):
    """Exception to be raised when current version is newer than the one to be processed."""

    def __init__(self, previous: Aggregate, new: Aggregate):
        self.previous = previous
        self.new = new
        super().__init__(
            f"Version for {repr(previous.classname)} aggregate must be "
            f"greater than {previous.version}. Obtained: {new.version}"
        )



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

from typing import (
    AsyncIterator,
    NoReturn,
)

from aiokafka import (
    AIOKafkaProducer,
)

from minos.common import (
    BROKER,
    QUEUE,
    MinosConfig,
)

from .abc import (
    BrokerSetup,
)


class Producer(BrokerSetup):
    """Minos Queue Dispatcher Class."""

    # noinspection PyUnresolvedReferences
    def __init__(self, *args, queue: QUEUE, broker: BROKER, **kwargs):
        # noinspection PyProtectedMember
        super().__init__(*args, **queue._asdict(), **kwargs)
        self.retry = queue.retry
        self.records = queue.records
        self.broker = broker

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> Producer:
        return cls(*args, **config.events._asdict(), **kwargs)

    async def dispatch(self) -> NoReturn:
        """Dispatch the items in the publishing queue.

        :return: This method does not return anything.
        """
        async for row in self.select():
            await self._dispatch_one(row)

    # noinspection PyUnusedLocal
    async def select(self, *args, **kwargs) -> AsyncIterator[tuple]:
        """Select a sequence of ``MinosSnapshotEntry`` objects.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A sequence of ``MinosSnapshotEntry`` objects.
        """
        params = (self.retry, self.records)
        async for row in self.submit_query_and_iter(_SELECT_NON_PROCESSED_ROWS_QUERY, params):
            yield row

    async def _dispatch_one(self, row: tuple) -> NoReturn:
        # noinspection PyBroadException
        published = await self.publish(topic=row[1], message=row[2])
        await self._update_queue_state(row, published)

    async def _update_queue_state(self, row: tuple, published: bool):
        update_query = _DELETE_PROCESSED_QUERY if published else _UPDATE_NON_PROCESSED_QUERY
        await self.submit_query(update_query, (row[0],))

    async def publish(self, topic: str, message: bytes) -> bool:
        """Publish a new item in in the broker (kafka).

        :param topic: The topic in which the message will be published.
        :param message: The message to be published.
        :return: A boolean flag, ``True`` when the message is properly published or ``False`` otherwise.
        """
        producer = AIOKafkaProducer(bootstrap_servers=f"{self.broker.host}:{self.broker.port}")
        # noinspection PyBroadException
        try:
            # Get cluster layout and initial topic/partition leadership information
            await producer.start()
            # Produce message
            await producer.send_and_wait(topic, message)
            published = True
        except Exception:
            published = False
        finally:
            # Wait for all pending messages to be delivered or expire.
            await producer.stop()

        return published


_SELECT_NON_PROCESSED_ROWS_QUERY = """
SELECT *
FROM producer_queue
WHERE retry <= %s
ORDER BY creation_date
LIMIT %s;
""".strip()

_DELETE_PROCESSED_QUERY = """
DELETE FROM producer_queue
WHERE id = %s;
""".strip()

_UPDATE_NON_PROCESSED_QUERY = """
UPDATE producer_queue
    SET retry = retry + 1
WHERE id = %s;
""".strip()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

from typing import (
    Any,
    AsyncIterator,
    NoReturn,
    Optional,
    Type,
)

from minos.common import (
    Aggregate,
    MinosConfig,
    MinosRepositoryAction,
    MinosRepositoryEntry,
    PostgreSqlMinosDatabase,
    PostgreSqlMinosRepository,
    import_module,
)

from ..exceptions import (
    MinosPreviousVersionSnapshotException,
)
from .entries import (
    SnapshotEntry,
)


class SnapshotBuilder(PostgreSqlMinosDatabase):
    """Minos Snapshot Dispatcher class."""

    def __init__(self, *args, repository: dict[str, Any] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.repository = PostgreSqlMinosRepository(**repository)

    async def _destroy(self) -> NoReturn:
        await super()._destroy()
        await self.repository.destroy()

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> SnapshotBuilder:
        return cls(*args, **config.snapshot._asdict(), repository=config.repository._asdict(), **kwargs)

    async def _setup(self) -> NoReturn:
        await self.submit_query(_CREATE_TABLE_QUERY)
        await self.submit_query(_CREATE_OFFSET_TABLE_QUERY)

    async def dispatch(self) -> NoReturn:
        """Perform a dispatching step, based on the sequence of non already processed ``MinosRepositoryEntry`` objects.

        :return: This method does not return anything.
        """
        offset = await self._load_offset()

        ids = set()

        def _update_offset(e: MinosRepositoryEntry, o: int):
            ids.add(e.id)
            while o + 1 in ids:
                o += 1
                ids.remove(o)
            return o

        async for entry in self.repository.select(id_ge=offset):
            try:
                await self._dispatch_one(entry)
            except MinosPreviousVersionSnapshotException:
                pass
            offset = _update_offset(entry, offset)

        await self._store_offset(offset)

    async def _load_offset(self) -> int:
        # noinspection PyBroadException
        try:
            raw = await self.submit_query_and_fetchone(_SELECT_OFFSET_QUERY)
            return raw[0]
        except Exception:
            return 0

    async def _store_offset(self, offset: int) -> NoReturn:
        await self.submit_query(_INSERT_OFFSET_QUERY, {"value": offset})

    # noinspection PyUnusedLocal
    async def select(self, *args, **kwargs) -> AsyncIterator[SnapshotEntry]:
        """Select a sequence of ``MinosSnapshotEntry`` objects.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A sequence of ``MinosSnapshotEntry`` objects.
        """
        async for row in self.submit_query_and_iter(_SELECT_ALL_ENTRIES_QUERY):
            yield SnapshotEntry(*row)

    async def _dispatch_one(self, event_entry: MinosRepositoryEntry) -> Optional[SnapshotEntry]:
        if event_entry.action is MinosRepositoryAction.DELETE:
            await self._submit_delete(event_entry)
            return
        instance = await self._build_instance(event_entry)
        return await self._submit_instance(instance)

    async def _submit_delete(self, entry: MinosRepositoryEntry) -> NoReturn:
        params = {"aggregate_id": entry.aggregate_id, "aggregate_name": entry.aggregate_name}
        await self.submit_query(_DELETE_ONE_SNAPSHOT_ENTRY_QUERY, params)

    async def _build_instance(self, event_entry: MinosRepositoryEntry) -> Aggregate:
        # noinspection PyTypeChecker
        cls: Type[Aggregate] = import_module(event_entry.aggregate_name)
        instance = cls.from_avro_bytes(event_entry.data, id=event_entry.aggregate_id, version=event_entry.version)
        instance = await self._update_if_exists(instance)
        return instance

    async def _update_if_exists(self, new: Aggregate) -> Aggregate:
        # noinspection PyBroadException
        try:
            # noinspection PyTypeChecker
            previous = await self._select_one_aggregate(new.id, new.classname)
        except Exception:
            return new

        if previous.version >= new.version:
            raise MinosPreviousVersionSnapshotException(previous, new)

        new._fields = previous.fields | new.fields
        return new

    async def _select_one_aggregate(self, aggregate_id: int, aggregate_name: str) -> Aggregate:
        snapshot_entry = await self._select_one(aggregate_id, aggregate_name)
        return snapshot_entry.aggregate

    async def _select_one(self, aggregate_id: int, aggregate_name: str) -> SnapshotEntry:
        raw = await self.submit_query_and_fetchone(_SELECT_ONE_SNAPSHOT_ENTRY_QUERY, (aggregate_id, aggregate_name))
        return SnapshotEntry(aggregate_id, aggregate_name, *raw)

    async def _submit_instance(self, aggregate: Aggregate) -> SnapshotEntry:
        snapshot_entry = SnapshotEntry.from_aggregate(aggregate)
        snapshot_entry = await self._submit_update_or_create(snapshot_entry)
        return snapshot_entry

    async def _submit_update_or_create(self, entry: SnapshotEntry) -> SnapshotEntry:
        params = {
            "aggregate_id": entry.aggregate_id,
            "aggregate_name": entry.aggregate_name,
            "version": entry.version,
            "data": entry.data,
        }
        response = await self.submit_query_and_fetchone(_INSERT_ONE_SNAPSHOT_ENTRY_QUERY, params)

        entry.created_at, entry.updated_at = response

        return entry


_CREATE_TABLE_QUERY = """
CREATE TABLE IF NOT EXISTS snapshot (
    aggregate_id BIGINT NOT NULL,
    aggregate_name TEXT NOT NULL,
    version INT NOT NULL,
    data BYTEA NOT NULL,
    created_at TIMESTAMPTZ NOT NULL DEFAULT NOW(),
    updated_at TIMESTAMPTZ NOT NULL DEFAULT NOW(),
    PRIMARY KEY (aggregate_id, aggregate_name)
);
""".strip()

_SELECT_ALL_ENTRIES_QUERY = """
SELECT aggregate_id, aggregate_name, version, data, created_at, updated_at
FROM snapshot;
""".strip()

_DELETE_ONE_SNAPSHOT_ENTRY_QUERY = """
DELETE FROM snapshot
WHERE aggregate_id = %(aggregate_id)s and aggregate_name = %(aggregate_name)s;
""".strip()

_SELECT_ONE_SNAPSHOT_ENTRY_QUERY = """
SELECT version, data, created_at, updated_at
FROM snapshot
WHERE aggregate_id = %s and aggregate_name = %s;
""".strip()

_INSERT_ONE_SNAPSHOT_ENTRY_QUERY = """
INSERT INTO snapshot (aggregate_id, aggregate_name, version, data, created_at, updated_at)
VALUES (
    %(aggregate_id)s,
    %(aggregate_name)s,
    %(version)s,
    %(data)s,
    default,
    default
)
ON CONFLICT (aggregate_id, aggregate_name)
DO
   UPDATE SET version = %(version)s, data = %(data)s, updated_at = NOW()
RETURNING created_at, updated_at;
""".strip()

_CREATE_OFFSET_TABLE_QUERY = """
CREATE TABLE IF NOT EXISTS snapshot_aux_offset (
    id bool PRIMARY KEY DEFAULT TRUE,
    value BIGINT NOT NULL,
    CONSTRAINT id_uni CHECK (id)
);
""".strip()

_SELECT_OFFSET_QUERY = """
SELECT value
FROM snapshot_aux_offset
WHERE id = TRUE;
"""
_INSERT_OFFSET_QUERY = """
INSERT INTO snapshot_aux_offset (id, value)
VALUES (TRUE, %(value)s)
ON CONFLICT (id)
DO UPDATE SET value = %(value)s;
""".strip()



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.
from __future__ import (
    annotations,
)

from typing import (
    Any,
    NoReturn,
)

from dependency_injector.wiring import (
    Provide,
)

from minos.common import (
    Command,
    MinosBroker,
    MinosConfig,
)

from ..abc import (
    Handler,
)
from ..entries import (
    HandlerEntry,
)


class CommandHandler(Handler):
    """Command Handler class."""

    TABLE_NAME = "command_queue"

    broker: MinosBroker = Provide["command_reply_broker"]

    def __init__(self, *, service_name: str, broker: MinosBroker = None, **kwargs: Any):
        super().__init__(broker_group_name=f"command_{service_name}", **kwargs)

        if broker is not None:
            self.broker = broker

    def _build_data(self, value: bytes) -> Command:
        return Command.from_avro_bytes(value)

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> CommandHandler:
        handlers = {item.name: {"controller": item.controller, "action": item.action} for item in config.commands.items}
        return cls(service_name=config.service.name, handlers=handlers, **config.commands.queue._asdict(), **kwargs)

    async def _dispatch_one(self, row: HandlerEntry) -> NoReturn:
        command: Command = row.data
        definition_id = command.saga_uuid

        response = await row.callback(row.topic, command)

        if command.reply_topic is not None:
            await self.broker.send(response, topic=command.reply_topic, saga_uuid=definition_id)



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.

from __future__ import (
    annotations,
)

from abc import (
    abstractmethod,
)
from datetime import (
    datetime,
)
from typing import (
    Any,
    Callable,
    NoReturn,
)

from minos.common import (
    MinosModel,
    import_module,
)
from minos.common.logs import (
    log,
)

from ...exceptions import (
    MinosNetworkException,
)
from ..entries import (
    HandlerEntry,
)
from .setups import (
    HandlerSetup,
)


class Handler(HandlerSetup):
    """
    Event Handler

    """

    __slots__ = "_handlers", "_handlers"

    def __init__(self, *, records: int, handlers: dict[str, dict[str, Any]], **kwargs: Any):
        super().__init__(**kwargs)
        self._handlers = handlers
        self._records = records

    async def dispatch(self) -> NoReturn:
        """Event Queue Checker and dispatcher.

        It is in charge of querying the database and calling the action according to the topic.

            1. Get periodically 10 records (or as many as defined in config > queue > records).
            2. Instantiate the action (asynchronous) by passing it the model.
            3. If the invoked function terminates successfully, remove the event from the database.

        Raises:
            Exception: An error occurred inserting record.
        """
        iterable = self.submit_query_and_iter(
            "SELECT * FROM %s ORDER BY creation_date ASC LIMIT %d;" % (self.TABLE_NAME, self._records),
        )
        async for row in iterable:
            try:
                await self.dispatch_one(row)
            except Exception as exc:
                log.warning(exc)
                continue
            await self.submit_query("DELETE FROM %s WHERE id=%d;" % (self.TABLE_NAME, row[0]))

    async def dispatch_one(self, row: tuple[int, str, int, bytes, datetime]) -> NoReturn:
        """Dispatch one row.

        :param row: Row to be dispatched.
        :return: This method does not return anything.
        """
        id = row[0]
        topic = row[1]
        callback = self.get_event_handler(row[1])
        partition_id = row[2]
        data = self._build_data(row[3])
        created_at = row[4]

        entry = HandlerEntry(id, topic, callback, partition_id, data, created_at)

        await self._dispatch_one(entry)

    def get_event_handler(self, topic: str) -> Callable:

        """Get Event instance to call.

        Gets the instance of the class and method to call.

        Args:
            topic: Kafka topic. Example: "TicketAdded"

        Raises:
            MinosNetworkException: topic TicketAdded have no controller/action configured, please review th
                configuration file.
        """
        if topic not in self._handlers:
            raise MinosNetworkException(
                f"topic {topic} have no controller/action configured, " f"please review th configuration file"
            )

        event = self._handlers[topic]
        # the topic exist, get the controller and the action
        controller = event["controller"]
        action = event["action"]

        object_class = import_module(controller)
        log.debug(object_class())
        instance_class = object_class()
        class_method = getattr(instance_class, action)

        return class_method

    @abstractmethod
    def _build_data(self, value: bytes) -> MinosModel:
        raise NotImplementedError

    @abstractmethod
    async def _dispatch_one(self, row: HandlerEntry) -> NoReturn:
        raise NotImplementedError



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from abc import (
    ABC,
)
from datetime import (
    datetime,
)
from typing import (
    NoReturn,
    Optional,
)

from minos.common import (
    MinosBroker,
    PostgreSqlMinosDatabase,
)


class BrokerSetup(PostgreSqlMinosDatabase):
    """Minos Broker Setup Class"""

    async def _setup(self) -> NoReturn:
        await self._create_broker_table()

    async def _create_broker_table(self) -> NoReturn:
        await self.submit_query(_CREATE_TABLE_QUERY)


class Broker(MinosBroker, BrokerSetup, ABC):
    """Minos Broker Class."""

    ACTION: str

    def __init__(self, topic: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.topic = topic

    async def _send_bytes(self, topic: str, raw: bytes) -> int:
        params = (topic, raw, 0, self.ACTION, datetime.now(), datetime.now())
        raw = await self.submit_query_and_fetchone(_INSERT_ENTRY_QUERY, params)
        return raw[0]


_CREATE_TABLE_QUERY = """
CREATE TABLE IF NOT EXISTS producer_queue (
    id BIGSERIAL NOT NULL PRIMARY KEY,
    topic VARCHAR(255) NOT NULL,
    model BYTEA NOT NULL,
    retry INTEGER NOT NULL,
    action VARCHAR(255) NOT NULL,
    creation_date TIMESTAMP NOT NULL,
    update_date TIMESTAMP NOT NULL
);
""".strip()

_INSERT_ENTRY_QUERY = """
INSERT INTO producer_queue (topic, model, retry, action, creation_date, update_date)
VALUES (%s, %s, %s, %s, %s, %s)
RETURNING id;
""".strip()



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.
from __future__ import (
    annotations,
)

from typing import (
    Any,
    NoReturn,
)

from minos.common import (
    Event,
    MinosConfig,
)

from ..abc import (
    Handler,
)
from ..entries import (
    HandlerEntry,
)


class EventHandler(Handler):
    """Event Handler class."""

    TABLE_NAME = "event_queue"

    def __init__(self, *, service_name: str, **kwargs: Any):
        super().__init__(broker_group_name=f"event_{service_name}", **kwargs)

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> EventHandler:
        handlers = {item.name: {"controller": item.controller, "action": item.action} for item in config.events.items}
        return cls(service_name=config.service.name, handlers=handlers, **config.events.queue._asdict(), **kwargs)

    def _build_data(self, value: bytes) -> Event:
        return Event.from_avro_bytes(value)

    async def _dispatch_one(self, row: HandlerEntry) -> NoReturn:
        await row.callback(row.topic, row.data)



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.
from __future__ import (
    annotations,
)

import functools
from typing import (
    Callable,
)

from aiohttp import (
    web,
)

from minos.common import (
    MinosConfig,
    MinosSetup,
    import_module,
)


class RestBuilder(MinosSetup):
    """
    Rest Interface Handler

    Rest Interface for aiohttp web handling.

    """

    __slots__ = "_config", "_app"

    def __init__(self, config: MinosConfig, app: web.Application = web.Application(), **kwargs):
        super().__init__(**kwargs)
        self._config = config
        self._app = app
        self.load_routes()

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> RestBuilder:
        return cls(*args, config=config, **kwargs)

    def load_routes(self):
        """Load routes from config file."""
        for item in self._config.rest.endpoints:
            callable_f = self.resolve_action(item.controller, item.action)
            self._app.router.add_route(item.method, item.route, callable_f)

        # Load default routes
        self._mount_system_health()

    def resolve_action(self, controller: str, action: str) -> Callable:
        """Load controller class and action method.
        :param controller: Controller string. Example: "tests.service.CommandTestService.CommandService"
        :param action: Config instance. Example: "get_order"
        :return: A class method callable instance.
        """
        object_class = import_module(controller)
        instance_class = object_class()
        class_method = getattr(instance_class, action)
        partial = functools.partial(class_method, config=self._config)

        return partial

    def get_app(self):
        """Return rest application instance.
        :return: A `web.Application` instance.
        """
        return self._app

    def _mount_system_health(self):
        """Mount System Health Route."""
        self._app.router.add_get("/system/health", self._system_health_handler)

    @staticmethod
    async def _system_health_handler(request):
        """System Health Route Handler.
        :return: A `web.json_response` response.
        """
        return web.json_response({"host": request.host}, status=200)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from .builders import (
    SnapshotBuilder,
)
from .entries import (
    SnapshotEntry,
)
from .services import (
    SnapshotService,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['minos',
 'minos.networks',
 'minos.networks.brokers',
 'minos.networks.handlers',
 'minos.networks.handlers.abc',
 'minos.networks.handlers.command_replies',
 'minos.networks.handlers.commands',
 'minos.networks.handlers.events',
 'minos.networks.rest',
 'minos.networks.snapshots']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0',
 'aiokafka>=0.7.0,<0.8.0',
 'aiomisc>=14.0.3,<15.0.0',
 'aiopg>=1.2.1,<2.0.0',
 'dependency-injector>=4.32.2,<5.0.0',
 'minos-microservice-common>=0.0,<0.1']

setup_kwargs = {
    'name': 'minos-microservice-networks',
    'version': '0.0.5',
    'description': 'Python Package with the common network classes and utilities used in Minos Microservice.',
    'long_description': "# Minos Microservice Network\n\n[![codecov](https://codecov.io/gh/Clariteia/minos_microservice_common/branch/main/graph/badge.svg)](https://codecov.io/gh/Clariteia/minos_microservice_common)\n![Tests](https://github.com/Clariteia/minos_microservice_common/actions/workflows/python-tests.yml/badge.svg)\n\nMinos is a framework which helps you create [reactive](https://www.reactivemanifesto.org/) microservices in Python.\nInternally, it leverages Event Sourcing, CQRS and a message driven architecture to fulfil the commitments of an\nasynchronous environment.\n\n## Set up a development environment\n\nMinos uses `poetry` as its default package manager. Please refer to\n(https://python-poetry.org/docs/#installation) for instructions on how to install it.\n\nNow you con install all the dependencies by running\n```bash\nmake install\n```\n\nIn order to make the pre-commits checks available to git, run\n```bash\npre-commit install\n```\n\nMake yourself sure you are able to run the tests. Refer to the appropriate section in this guide.\n\n## Run the tests\n\nIn order to run the tests, please make sure you have the [Docker Engine](https://docs.docker.com/engine/install/)\nand [Docker Compose](https://docs.docker.com/compose/install/) installed.\n\nMove into `tests/` directory\n\n```bash\ncd tests/\n```\nRun service dependencies:\n\n```bash\ndocker-compose up -d\n```\n\nInstall library dependencies:\n\n```bash\nmake install\n```\n\nRun tests:\n\n```bash\nmake test\n```\n\n## How to contribute\n\nMinos being an open-source project, we are looking forward to having your contributions. No matter whether it is a pull\nrequest with new features, or the creation of an issue related to a bug you have found.\n\nPlease consider these guidelines before you submit any modification.\n\n### Create an issue\n\n1. If you happen to find a bug, please file a new issue filling the 'Bug report' template.\n2. Set the appropriate labels, so we can categorise it easily.\n3. Wait for any core developer's feedback on it.\n\n### Submit a Pull Request\n\n1. Create an issue following the previous steps.\n2. Fork the project.\n3. Push your changes to a local branch.\n4. Run the tests!\n5. Submit a pull request from your fork's branch.\n\n## Credits\n\nThis package was created with ![Cookiecutter](https://github.com/audreyr/cookiecutter) and the ![Minos Package](https://github.com/Clariteia/minos-pypackage) project template.\n\n",
    'author': 'Clariteia Devs',
    'author_email': 'devs@clariteia.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://clariteia.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

from aiomisc.service.aiohttp import (
    AIOHTTPService,
)

from minos.common import (
    MinosConfig,
)

from .builders import (
    RestBuilder,
)


class RestService(AIOHTTPService):
    """
    Rest Interface

    Expose REST Interface handler using aiomisc AIOHTTPService.

    """

    def __init__(self, config: MinosConfig, **kwargs):
        address = config.rest.broker.host
        port = config.rest.broker.port
        super().__init__(address=address, port=port, **kwargs)
        self.rest = RestBuilder.from_config(config=config, **kwargs)

    async def create_application(self):
        return self.rest.get_app()  # pragma: no cover



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from .abc import (
    Consumer,
    Handler,
    HandlerSetup,
)
from .command_replies import (
    CommandReplyConsumer,
    CommandReplyConsumerService,
    CommandReplyHandler,
    CommandReplyHandlerService,
)
from .commands import (
    CommandConsumer,
    CommandConsumerService,
    CommandHandler,
    CommandHandlerService,
)
from .entries import (
    HandlerEntry,
)
from .events import (
    EventConsumer,
    EventConsumerService,
    EventHandler,
    EventHandlerService,
)



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.
from __future__ import (
    annotations,
)

from typing import (
    Any,
    NoReturn,
)

from dependency_injector.wiring import (
    Provide,
)

from minos.common import (
    CommandReply,
    MinosConfig,
    MinosSagaManager,
)

from ..abc import (
    Handler,
)
from ..entries import (
    HandlerEntry,
)


class CommandReplyHandler(Handler):
    """Command Reply Handler class."""

    TABLE_NAME = "command_reply_queue"

    saga_manager: MinosSagaManager = Provide["saga_manager"]

    def __init__(self, *, service_name: str, saga_manager: MinosSagaManager = None, **kwargs: Any):
        super().__init__(broker_group_name=f"command_reply_{service_name}", **kwargs)

        if saga_manager is not None:
            self.saga_manager = saga_manager

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> CommandReplyHandler:
        handlers = {
            f"{item.name}Reply": {"controller": item.controller, "action": item.action} for item in config.saga.items
        }
        return cls(*args, service_name=config.service.name, handlers=handlers, **config.saga.queue._asdict(), **kwargs,)

    def _build_data(self, value: bytes) -> CommandReply:
        return CommandReply.from_avro_bytes(value)

    async def _dispatch_one(self, row: HandlerEntry) -> NoReturn:
        await self.saga_manager.run(reply=row.data)
